#include "ColorDefs.h"

namespace vivid {
    extern const std::map<std::string, color_t> COLORS {
        {"indianred", INDIANRED},
        {"lightcoral", LIGHTCORAL},
        {"salmon", SALMON},
        {"darksalmon", DARKSALMON},
        {"lightsalmon", LIGHTSALMON},
        {"crimson", CRIMSON},
        {"red", RED},
        {"darkred", DARKRED},
        {"pink", PINK},
        {"lightpink", LIGHTPINK},
        {"hotpink", HOTPINK},
        {"deeppink", DEEPPINK},
        {"mediumvioletred", MEDIUMVIOLETRED},
        {"palevioletred", PALEVIOLETRED},
        {"coral", CORAL},
        {"tomato", TOMATO},
        {"orangered", ORANGERED},
        {"darkorange", DARKORANGE},
        {"orange", ORANGE},
        {"gold", GOLD},
        {"yellow", YELLOW},
        {"lightyellow", LIGHTYELLOW},
        {"lemonchiffon", LEMONCHIFFON},
        {"lightgoldenrodyellow", LIGHTGOLDENRODYELLOW},
        {"papayawhip", PAPAYAWHIP},
        {"moccasin", MOCCASIN},
        {"peachpuff", PEACHPUFF},
        {"palegoldenrod", PALEGOLDENROD},
        {"khaki", KHAKI},
        {"darkkhaki", DARKKHAKI},
        {"lavender", LAVENDER},
        {"thistle", THISTLE},
        {"plum", PLUM},
        {"violet", VIOLET},
        {"orchid", ORCHID},
        {"fuchsia", FUCHSIA},
        {"magenta", MAGENTA},
        {"mediumorchid", MEDIUMORCHID},
        {"mediumpurple", MEDIUMPURPLE},
        {"rebeccapurple", REBECCAPURPLE},
        {"blueviolet", BLUEVIOLET},
        {"darkviolet", DARKVIOLET},
        {"darkorchid", DARKORCHID},
        {"darkmagenta", DARKMAGENTA},
        {"purple", PURPLE},
        {"indigo", INDIGO},
        {"slateblue", SLATEBLUE},
        {"darkslateblue", DARKSLATEBLUE},
        {"mediumslateblue", MEDIUMSLATEBLUE},
        {"greenyellow", GREENYELLOW},
        {"chartreuse", CHARTREUSE},
        {"lawngreen", LAWNGREEN},
        {"lime", LIME},
        {"limegreen", LIMEGREEN},
        {"palegreen", PALEGREEN},
        {"lightgreen", LIGHTGREEN},
        {"mediumspringgreen", MEDIUMSPRINGGREEN},
        {"springgreen", SPRINGGREEN},
        {"mediumseagreen", MEDIUMSEAGREEN},
        {"seagreen", SEAGREEN},
        {"forestgreen", FORESTGREEN},
        {"green", GREEN},
        {"darkgreen", DARKGREEN},
        {"yellowgreen", YELLOWGREEN},
        {"olivedrab", OLIVEDRAB},
        {"olive", OLIVE},
        {"darkolivegreen", DARKOLIVEGREEN},
        {"mediumaquamarine", MEDIUMAQUAMARINE},
        {"darkseagreen", DARKSEAGREEN},
        {"lightseagreen", LIGHTSEAGREEN},
        {"darkcyan", DARKCYAN},
        {"teal", TEAL},
        {"aqua", AQUA},
        {"cyan", CYAN},
        {"lightcyan", LIGHTCYAN},
        {"paleturquoise", PALETURQUOISE},
        {"aquamarine", AQUAMARINE},
        {"turquoise", TURQUOISE},
        {"mediumturquoise", MEDIUMTURQUOISE},
        {"darkturquoise", DARKTURQUOISE},
        {"cadetblue", CADETBLUE},
        {"steelblue", STEELBLUE},
        {"lightsteelblue", LIGHTSTEELBLUE},
        {"powderblue", POWDERBLUE},
        {"lightblue", LIGHTBLUE},
        {"skyblue", SKYBLUE},
        {"lightskyblue", LIGHTSKYBLUE},
        {"deepskyblue", DEEPSKYBLUE},
        {"dodgerblue", DODGERBLUE},
        {"cornflowerblue", CORNFLOWERBLUE},
        {"royalblue", ROYALBLUE},
        {"blue", BLUE},
        {"mediumblue", MEDIUMBLUE},
        {"darkblue", DARKBLUE},
        {"navy", NAVY},
        {"midnightblue", MIDNIGHTBLUE},
        {"cornsilk", CORNSILK},
        {"blanchedalmond", BLANCHEDALMOND},
        {"bisque", BISQUE},
        {"navajowhite", NAVAJOWHITE},
        {"wheat", WHEAT},
        {"burlywood", BURLYWOOD},
        {"tan", TAN},
        {"rosybrown", ROSYBROWN},
        {"sandybrown", SANDYBROWN},
        {"goldenrod", GOLDENROD},
        {"darkgoldenrod", DARKGOLDENROD},
        {"peru", PERU},
        {"chocolate", CHOCOLATE},
        {"saddlebrown", SADDLEBROWN},
        {"sienna", SIENNA},
        {"brown", BROWN},
        {"maroon", MAROON},
        {"white", WHITE},
        {"snow", SNOW},
        {"honeydew", HONEYDEW},
        {"mintcream", MINTCREAM},
        {"azure", AZURE},
        {"aliceblue", ALICEBLUE},
        {"ghostwhite", GHOSTWHITE},
        {"whitesmoke", WHITESMOKE},
        {"seashell", SEASHELL},
        {"beige", BEIGE},
        {"oldlace", OLDLACE},
        {"floralwhite", FLORALWHITE},
        {"ivory", IVORY},
        {"antiquewhite", ANTIQUEWHITE},
        {"linen", LINEN},
        {"lavenderblush", LAVENDERBLUSH},
        {"mistyrose", MISTYROSE},
        {"gainsboro", GAINSBORO},
        {"lightgray", LIGHTGRAY},
        {"silver", SILVER},
        {"darkgray", DARKGRAY},
        {"gray", GRAY},
        {"dimgray", DIMGRAY},
        {"lightslategray", LIGHTSLATEGRAY},
        {"slategray", SLATEGRAY},
        {"darkslategray", DARKSLATEGRAY},
        {"black", BLACK},
    };
}