#include "revision.h"
#if defined(__GNUC__) && defined(_WIN32)
	#include "winresrc.h"
#else
	#include "winres.h"
#endif

LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

IDI_ASSIMP_CMD         ICON                    "../shared/assimp_tools_icon.ico"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION VER_FILEVERSION
 PRODUCTVERSION VER_FILEVERSION
 FILEFLAGSMASK VS_FF_DEBUG
#ifdef _DEBUG
 FILEFLAGS VS_FF_DEBUG
#endif
 FILEOS VOS_NT
 FILETYPE VFT_APP
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904B0"
        BEGIN
            VALUE "Comments", "Licensed under a 3-clause BSD license"
            VALUE "CompanyName", "ASSIMP Team"
            VALUE "FileDescription", "ASSIMP CMD Application"
            VALUE "FileVersion", VER_FILEVERSION_STR
            VALUE "InternalName", "assimp"
            VALUE "LegalCopyright", VER_COPYRIGHT_STR
            VALUE "OriginalFilename", "assimp.exe"
            VALUE "ProductName", "ASSIMP CMD Application"
            VALUE "ProductVersion", VER_FILEVERSION_STR
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 65001
    END
END

