/*! \file int2str.hpp
  \brief A function that converts an integer to a string
  \author Almog Yalinewich
*/

#ifndef INT2STR_HPP
#define INT2STR_HPP 1

#include <string>

/*! \brief Converts an integer to a string
 */
std::string int2str(int n);

#endif
