#ifndef VIVID_MARCHINGCUBESSURFACE_H
#define VIVID_MARCHINGCUBESSURFACE_H

namespace vivid
{

/* Implementing surfacing by MarchingCubes algorithm */
class CMarchingCubesSurface {

};

}; // namespace vivid
#endif //VIVID_MARCHINGCUBESSURFACE_H
