declare -a headers=("fast_atof.h"
  "qnan.h"
  "BaseImporter.h"
  "Hash.h"
  "MemoryIOWrapper.h"
  "ParsingUtils.h"
  "StreamReader.h"
  "StreamWriter.h"
  "StringComparison.h"
  "StringUtils.h"
  "SGSpatialSort.h"
  "GenericProperty.h"
  "SpatialSort.h"
  "SkeletonMeshBuilder.h"
  "SmoothingGroups.h"
  "StandardShapes.h"
  "RemoveComments.h"
  "Subdivision.h"
  "Vertex.h"
  "LineSplitter.h"
  "TinyFormatter.h"
  "Profiler.h"
  "LogAux.h"
  "Bitmap.h"
  "XMLTools.h"
  "IOStreamBuffer.h"
  "CreateAnimMesh.h"
  "irrXMLWrapper.h"
  "BlobIOSystem.h"
  "MathFunctions.h"
  "Macros.h"
  "Exceptional.h"
  "ByteSwapper.h")

for i in "${headers[@]}"
do
  find . -type f -exec sed -i "s,<../code/$i>,<assimp/$i>,g" {} \;
done
