mod anim;
/* Animation
 * NodeAnim
 * MeshAnim
 * MeshMorphAnim
 */
mod blob;
/* ExportDataBlob
 */
mod vec;
/* Vector2d
 * Vector3d
 * */
mod matrix;
/* Matrix3by3
 * Matrix4by4
 */
mod camera;
/* Camera */
mod color;
/* Color3d
 * Color4d
 */
mod key;
/* MeshKey
 * MeshMorphKey
 * QuatKey
 * VectorKey
 */
mod texel;
mod plane;
mod string;
/* String
 */
mod material;
/* Material
 * MaterialPropery
 * MaterialPropertyString
 */
mod mem;
mod quaternion;
mod face;
mod vertex_weight;
mod mesh;
/* Mesh
 */
mod meta;
/* Metadata
 * MetadataEntry
 */
mod node;
/* Node
 * */
mod light;
mod texture;
mod ray;
mod transform;
/* UVTransform */
mod bone;
mod scene;
/* Scene */
