FORMATS = ["CSM", 
            "LWS", 
            "B3D", 
            "COB", 
            "PLY", 
            "IFC", 
            "OFF", 
            "SMD", 
            "IRRMESH", 
            "3D", 
            "DAE", 
            "MDL", 
            "HMP", 
            "TER", 
            "WRL", 
            "XML", 
            "NFF", 
            "AC", 
            "OBJ", 
            "3DS", 
            "STL", 
            "IRR", 
            "Q3O",
            "Q3D",
            "MS3D", 
            "Q3S", 
            "ZGL", 
            "MD2", 
            "X", 
            "BLEND", 
            "XGL", 
            "MD5MESH", 
            "MAX", 
            "LXO", 
            "DXF", 
            "BVH", 
            "LWO",
            "NDO"]

def available_formats():
    return FORMATS
