// Microsoft Visual C++ generated resource script.
//
#include "resource.h"
#include "revision.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#define APSTUDIO_HIDDEN_SYMBOLS
#include "windows.h"
#undef APSTUDIO_HIDDEN_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ASSIMP_VIEW         ICON                    "../shared/assimp_tools_icon.ico"


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOGEX 22, 17, 283, 149
STYLE DS_SETFONT | DS_MODALFRAME | DS_CENTER | WS_CAPTION | WS_SYSMENU
CAPTION "About Open Asset Import Library"
FONT 9, "Courier New", 400, 0, 0x0
BEGIN
    LTEXT           "Open Asset Import Library (Assimp)",IDC_STATIC,30,14,144,9
    LTEXT           "A free C/C++ library to read various well-known 3D model formats into a straightforward in-memory format for easy processing by applications. Licensed under a 3-clause BSD license and totally awesome.",IDC_STATIC,31,34,204,24
    LTEXT           "\xA9 2008-2022. Assimp Development Team. See the CREDITS file for a list of all contributors.",IDC_STATIC,30,65,204,23
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,27,282,1
    LTEXT           "https://assimp.org",IDC_STATIC,31,101,127,22
    DEFPUSHBUTTON   "Love this library",IDOK,186,110,84,14
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,148,283,1
    CONTROL         IDB_BITMAP1,IDC_STATIC,"Static",SS_BITMAP,0,129,514,20
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,10,281,1
END

IDD_DIALOGMAIN DIALOGEX 0, 0, 656, 450
STYLE DS_SETFONT | DS_MODALFRAME | DS_CENTER | WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_POPUP | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_ACCEPTFILES | WS_EX_WINDOWEDGE
CAPTION "Open Asset Import Library - Model Viewer "
MENU IDR_MENU1
FONT 8, "Microsoft Sans Serif", 400, 0, 0x0
BEGIN
    CONTROL         "",IDC_RT,"Static",SS_OWNERDRAW,0,0,513,324
    CONTROL         "",IDC_TREE1,"SysTreeView32",TVS_HASBUTTONS | TVS_HASLINES | TVS_SHOWSELALWAYS | WS_BORDER | WS_HSCROLL | WS_TABSTOP,513,0,143,450
    CONTROL         "<<",IDC_BLUBB,"Button",BS_AUTOCHECKBOX | BS_PUSHLIKE | WS_TABSTOP,471,328,35,14
    COMBOBOX        IDC_COMBO1,367,328,100,14,CBS_DROPDOWN | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Play",IDC_PLAY,328,328,35,14
    CONTROL         "",IDC_SLIDERANIM,"msctls_trackbar32",TBS_AUTOTICKS | TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,0,328,328,15
    GROUPBOX        "Display",IDC_STATIC,7,345,172,101
    CONTROL         "Multisampling [M]",IDC_TOGGLEMS,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,357,80,10
    CONTROL         "Wireframe [W]",IDC_TOGGLEWIRE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,369,80,10
    CONTROL         "No materials [D]",IDC_TOGGLEMAT,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,381,80,10
    CONTROL         "Display normals [N]",IDC_TOGGLENORMALS,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,393,80,10
    CONTROL         "Low quality [P]",IDC_LOWQUALITY,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,405,80,10
    CONTROL         "No specular [S]",IDC_NOSPECULAR,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,417,80,10
    CONTROL         "Show skeleton [K]",IDC_SHOWSKELETON,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,13,429,80,10
    CONTROL         "AutoRotate [A]",IDC_AUTOROTATE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,93,357,80,10
    CONTROL         "Zoom/Rotate [Z]",IDC_ZOOM,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,93,369,80,10
    CONTROL         "Rotate lights [R]",IDC_LIGHTROTATE,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,93,381,80,10
    CONTROL         "Two lights [L]",IDC_3LIGHTS,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,93,393,80,10
    CONTROL         "Backface culling [C]",IDC_BFCULL,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,93,405,80,10
    CONTROL         "No transparency [T]",IDC_NOAB,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,93,417,80,10
    GROUPBOX        "Statistics",IDC_STATIC,186,345,164,63
    LTEXT           "Vertices:",IDC_NUMVERTS,192,357,35,8
    LTEXT           "Nodes:",IDC_NUMNODES,192,369,35,8
    LTEXT           "Shaders:",IDC_NUMSHADERS,192,381,35,8
    LTEXT           "Time:",IDC_LOADTIME,192,393,35,8
    EDITTEXT        IDC_EVERT,227,357,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_ENODEWND,227,369,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_ESHADER,227,381,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_ELOAD,227,393,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    LTEXT           "Faces:",IDC_NUMFACES,272,357,35,8
    LTEXT           "Materials:",IDC_NUMMATS,272,369,35,8
    LTEXT           "Meshes:",IDC_NUMMESHES,272,381,35,8
    LTEXT           "FPS:",IDC_FPS,272,393,35,8
    EDITTEXT        IDC_EFACE,307,357,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_EMAT,307,369,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_EMESH,307,381,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_EFPS,307,393,35,8,ES_RIGHT | ES_AUTOHSCROLL | ES_READONLY | NOT WS_BORDER
    EDITTEXT        IDC_VIEWMATRIX,192,412,72,44,ES_MULTILINE | ES_AUTOHSCROLL | ES_READONLY | NOT WS_VISIBLE | NOT WS_BORDER
    GROUPBOX        "Colors",IDC_STATIC,357,345,109,87
    LTEXT           "Primary:",IDC_STATIC,363,360,48,8
    LTEXT           "Secondary:",IDC_STATIC,363,378,54,8
    LTEXT           "Ambient:",IDC_STATIC,363,396,54,8
    CONTROL         "Button1",IDC_LCOLOR1,"Button",BS_OWNERDRAW | WS_TABSTOP,423,357,35,14
    CONTROL         "Button1",IDC_LCOLOR2,"Button",BS_OWNERDRAW | WS_TABSTOP,423,375,35,14
    CONTROL         "Button1",IDC_LCOLOR3,"Button",BS_OWNERDRAW | WS_TABSTOP,423,393,35,14
    PUSHBUTTON      "Reset",IDC_LRESET,423,411,35,14
END

IDD_LOADDIALOG DIALOGEX 0, 0, 143, 60
STYLE DS_SETFONT | DS_CENTER | WS_POPUP | WS_BORDER | WS_SYSMENU
FONT 12, "Tahoma", 400, 0, 0x0
BEGIN
    DEFPUSHBUTTON   "Cancel",IDOK,104,41,33,13
    CONTROL         "",IDC_PROGRESS,"msctls_progress32",WS_BORDER,6,30,130,8
    LTEXT           "Loading ...\nMay the force be with you ...",IDC_STATIC,8,9,123,16
END

IDD_AVHELP DIALOGEX 0, 0, 481, 346
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "ASSIMP Viewer: Help"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,420,324,50,14
    CONTROL         "",IDC_RICHEDIT21,"RichEdit20A",ES_MULTILINE | ES_AUTOVSCROLL | ES_READONLY | WS_VSCROLL | WS_TABSTOP,19,18,462,294
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,312,490,1
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,16,490,1
END

IDD_LOGVIEW DIALOGEX 0, 0, 365, 182
STYLE DS_SETFONT | DS_CENTER | WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME
EXSTYLE WS_EX_TOPMOST | WS_EX_WINDOWEDGE
CAPTION "AssimpView Log Output"
FONT 8, "Courier New", 400, 0, 0x0
BEGIN
    CONTROL         "",IDC_EDIT1,"RichEdit20A",ES_MULTILINE | ES_AUTOVSCROLL | ES_READONLY | ES_NUMBER | WS_VSCROLL | WS_TABSTOP,3,4,358,174,WS_EX_STATICEDGE
END

IDD_DIALOGSMOOTH DIALOGEX 0, 0, 278, 141
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Set smooth limit "
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,213,94,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,161,94,50,14
    EDITTEXT        IDC_EDITSM,99,7,175,14,ES_AUTOHSCROLL | ES_NUMBER
    LTEXT           "Angle limit (in degrees):",IDC_STATIC,13,10,76,8
    LTEXT           "The angle limit defines the maximum angle that may be between two adjacent face normals that they're smoothed together.",IDC_STATIC,13,31,253,19
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,113,278,1
    LTEXT           "Also used during import, but can be overridden by single model importers to match the original look of a model as close as possible. Examples includes 3DS, ASE and LWO, all of them relying on smoothing groups and their own angle limits. ",IDC_STATIC,13,51,254,33
    LTEXT           "NOTE: New settings don't take effect immediately, use 'Smooth Normals' or 'Reload' to update the model.",IDC_STATIC,14,118,254,22
    CONTROL         "",IDC_STATIC,"Static",SS_ETCHEDHORZ,0,90,277,1
END


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#define APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "#include ""windows.h""\r\n"
    "#undef APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION VER_FILEVERSION
 PRODUCTVERSION VER_FILEVERSION
 FILEFLAGSMASK VS_FF_DEBUG
#ifdef _DEBUG
 FILEFLAGS VS_FF_DEBUG
#endif
 FILEOS VOS_NT
 FILETYPE VFT_APP
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904B0"
        BEGIN
            VALUE "Comments", "Licensed under a 3-clause BSD license"
            VALUE "CompanyName", "ASSIMP Team"
            VALUE "FileDescription", "ASSIMP Viewer Application"
            VALUE "FileVersion", VER_FILEVERSION_STR
            VALUE "InternalName", "assimp_view"
            VALUE "LegalCopyright", VER_COPYRIGHT_STR
            VALUE "OriginalFilename", "assimp_viewer.exe"
            VALUE "ProductName", "ASSIMP Viewer Application"
            VALUE "ProductVersion", VER_FILEVERSION_STR
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 65001
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        TOPMARGIN, 1
        BOTTOMMARGIN, 138
    END

    IDD_DIALOGMAIN, DIALOG
    BEGIN
        RIGHTMARGIN, 623
    END

    IDD_LOADDIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        TOPMARGIN, 7
    END

    IDD_AVHELP, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 474
        TOPMARGIN, 7
        BOTTOMMARGIN, 339
    END

    IDD_LOGVIEW, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 358
        TOPMARGIN, 14
        BOTTOMMARGIN, 175
    END

    IDD_DIALOGSMOOTH, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 271
        TOPMARGIN, 7
        BOTTOMMARGIN, 134
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_BITMAP1             BITMAP                  "banner.bmp"

IDB_BANIM               BITMAP                  "base_anim.bmp"

IDB_BDISPLAY            BITMAP                  "base_display.bmp"

IDB_BINTER              BITMAP                  "base_inter.bmp"

IDB_BRENDERING          BITMAP                  "base_rendering.bmp"

IDB_BSTATS              BITMAP                  "base_stats.bmp"

IDB_BTX                 BITMAP                  "tx.bmp"

IDB_BFX                 BITMAP                  "fx.bmp"

IDB_BNODE               BITMAP                  "n.bmp"

IDB_BROOT               BITMAP                  "root.bmp"

IDB_BTXI                BITMAP                  "txi.bmp"


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDR_MENU1 MENU
BEGIN
    POPUP "Viewer"
    BEGIN
        MENUITEM "Open Asset",                  ID_VIEWER_OPEN
        MENUITEM "Close Asset",                 ID_VIEWER_CLOSEASSET
        MENUITEM "Reload",                      ID_VIEWER_RELOAD
        POPUP "Import settings"
        BEGIN
            MENUITEM "Calculate Tangent Space",     ID_VIEWER_PP_CTS
            MENUITEM "Compute Indexed Meshes",      ID_VIEWER_PP_JIV
            MENUITEM "Optimize Materials",          ID_VIEWER_PP_RRM2
            MENUITEM "Optimize Meshes",             ID_VIEWER_PP_OM
            MENUITEM "Optimize Scenegraph",         ID_VIEWER_PP_OG
            MENUITEM "Find Instanced Meshes",       ID_VIEWER_PP_FIM
            MENUITEM "Run Full Validation",         ID_VIEWER_PP_VDS
            MENUITEM "Pretransform Vertices",       ID_VIEWER_PP_PTV
            MENUITEM "VCache Optimization",         ID_VIEWER_PP_ICL
            MENUITEM "Fix Infacing Normals",        ID_VIEWER_PP_FIN
            MENUITEM "Find Degenerates",            ID_VIEWER_PP_FD
            MENUITEM "Find Invalid Data",           ID_VIEWER_PP_FID
            MENUITEM "Generate UV Coords",          ID_VIEWER_PP_GUV
            MENUITEM "Transform UV Coords",         ID_VIEWER_PP_TUV
            MENUITEM "Remove Lines and Points",     ID_VIEWER_PP_RLINE_PNT, GRAYED
            MENUITEM "Remove dummy bones (De-bone)", ID_VIEWER_PP_DB
            MENUITEM SEPARATOR
            MENUITEM "(required) Triangulate",      ID_VIEWER_PP_JIV, GRAYED
            MENUITEM "(required) Limit Bone Weights", ID_VIEWER_PP_JIV, GRAYED
            MENUITEM "(required) Split Large Meshes", ID_VIEWER_PP_JIV, GRAYED
            MENUITEM "(required) Sort by primitive type", ID_VIEWER_PP_JIV, GRAYED
            MENUITEM "(required) Convert to Left-Handed", ID_VIEWER_PP_JIV, GRAYED
            MENUITEM SEPARATOR
            MENUITEM "Reset to default",            ID_IMPORTSETTINGS_RESETTODEFAULT
            MENUITEM "Open Post-Process Short Reference", ID_IMPORTSETTINGS_OPENPOST
        END
        MENUITEM SEPARATOR
        MENUITEM "Save Screenshot",             ID_VIEWER_SAVESCREENSHOTTOFILE
        MENUITEM "Reset view",                  ID_VIEWER_RESETVIEW
        MENUITEM "Memory consumption",          ID_VIEWER_MEMORYCONSUMATION
        MENUITEM SEPARATOR
        MENUITEM "Setup file associations",     ID_VIEWER_H
        MENUITEM SEPARATOR
        MENUITEM "Recent files ",               ID_VIEWER_RECENTFILES
        MENUITEM "Clear history",               ID_VIEWER_CLEARHISTORY
        MENUITEM SEPARATOR
        MENUITEM "Quit",                        ID_VIEWER_QUIT
    END
    POPUP "Tools"
    BEGIN
        MENUITEM "Log window",                  ID_TOOLS_LOGWINDOW
        MENUITEM "Save log to file",            ID_TOOLS_SAVELOGTOFILE
        MENUITEM "Clear log",                   ID_TOOLS_CLEARLOG
        MENUITEM SEPARATOR
        MENUITEM "Original normals",            ID_TOOLS_ORIGINALNORMALS, CHECKED
        MENUITEM "Hard normals",                ID_TOOLS_HARDNORMALS
        MENUITEM "Smooth normals",              ID_TOOLS_SMOOTHNORMALS
        MENUITEM SEPARATOR
        MENUITEM "Set angle limit ...",         ID_TOOLS_SETANGLELIMIT
        MENUITEM "Flip normals",                ID_TOOLS_FLIPNORMALS
        MENUITEM SEPARATOR
        MENUITEM "Stereo view",                 ID_TOOLS_STEREOVIEW
    END
    POPUP "Background"
    BEGIN
        MENUITEM "Set color",                   ID_BACKGROUND_SETCOLOR
        MENUITEM "Load skybox",                 ID_BACKGROUND_LOADSKYBOX
        MENUITEM "Load texture",                ID_BACKGROUND_LOADTEXTURE
        MENUITEM SEPARATOR
        MENUITEM "Clear",                       ID_BACKGROUND_CLEAR
    END
    MENUITEM "Export",                      32878
    POPUP "?"
    BEGIN
        POPUP "Feedback"
        BEGIN
            MENUITEM "Report bug",              ID_REPORTBUG
            MENUITEM "Feature request",         ID_FR
        END
        MENUITEM "Help",                        ID__HELP
        MENUITEM SEPARATOR
        MENUITEM "About",                       ID__ABOUT
        MENUITEM SEPARATOR
        MENUITEM "Website",                     ID__WEBSITE
        MENUITEM "GitHub Repository",           ID__WEBSITEGH
    END
END

IDR_TXPOPUP MENU
BEGIN
    POPUP "Hey"
    BEGIN
        MENUITEM "Replace texture",             ID_HEY_REPLACE
        MENUITEM "Export texture",              ID_HEY_EXPORT
        MENUITEM "Remove texture",              ID_HEY_REMOVE
        MENUITEM SEPARATOR
        MENUITEM "Reset texture",               ID_HEY_RESETTEXTURE
    END
    MENUITEM "This is not an easter egg",   0
END

IDR_MATPOPUP MENU
BEGIN
    POPUP "So long"
    BEGIN
        MENUITEM "Add diffuse texture",         ID_SOLONG_ADDDIFFUSETEXTURE
        MENUITEM "Add specular texture",        ID_SOLONG_ADDSPECULARTEXTURE
        MENUITEM "Add ambient texture",         ID_SOLONG_ADDAMBIENTTEXTURE
        MENUITEM "Add emissive texture",        ID_SOLONG_ADDEMISSIVETEXTURE
        MENUITEM "Add opacity texture",         ID_SOLONG_ADDOPACITYTEXTURE
        MENUITEM "Add normal/height texture",   ID_SOLONG_ADDNORMAL
        MENUITEM "Add shininess texture",       ID_SOLONG_ADDSHININESSTEXTURE
        MENUITEM SEPARATOR
        MENUITEM "Set diffuse color",           ID_SOLONG_CLEARDIFFUSECOLOR
        MENUITEM "Set specular color",          ID_SOLONG_CLEARSPECULARCOLOR
        MENUITEM "Set ambient color",           ID_SOLONG_CLEARAMBIENTCOLOR
        MENUITEM "Set emissive color",          ID_SOLONG_CLEAREMISSIVECOLOR
        MENUITEM "Set transparency",            ID_SOLONG_CLEARTRANSPARENCY
        MENUITEM SEPARATOR
        MENUITEM "Make default material",       ID_SOLONG_MAKEDEFAULTMATERIAL
        POPUP "Set shading mode"
        BEGIN
            MENUITEM "Gouraud",                     ID_SETSHADINGMODE_GOURAUD
            MENUITEM "Phong (specular)",            ID_SETSHADINGMODE_PHONG
        END
    END
    MENUITEM "and thanks for all the fish", 0
END


/////////////////////////////////////////////////////////////////////////////
//
// TEXT
//

IDR_TEXT1               TEXT                    "text1.bin"


/////////////////////////////////////////////////////////////////////////////
//
// RCDATA
//

IDR_HUD                 RCDATA                  "HUD.png"

IDR_HUDMASK             RCDATA                  "HUDMask.png"

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

